# Tutorial: Use `armarx_control`

[[_TOC_]]

---

to use the armarx_control feature, you need the following packages at the moment, it will
be cleaner soon.

- This package in `master` branch
- `armarx/skills/control` package (`main` branch) is where the NJointControllers are implemented.

## Using the Real Robot and `armarx_control` Module

### Example of using task space impedance controller:

1. Start ArmarX on the planning PC (e.g. `armar6a-1` on ARMAR-6) and start low/high-level scenarios.
   ```shell
   # In a new terminal on armar6a-1
   armarx start
   armarx memory start
   armarx gui
   # In the GUI, start low/high-level scenario
   ```
2. To use the Python binding on the robot with the new controllers, you should make sure the `armarx/skills/control`
   package is compiled on the control PC of the robot (e.g. `armar6a-0` on ARMAR-6).

   If you do not have your own workspace, you can use the following workspace if you do not change anything
    ```shell
    # In a new terminal on armar6a-0
    axii w act your_workspace
    start_unit
    # In a new terminal on armar6a-0
    axii w act your_workspace
    armarx gui
    # In the pop up GUI, find the component `controller_creator` in the `ControlScenario` scenario.
    # start the component, make sure you see "controller_creator is ready" in the log
    # Note, the `ControlScenario` is in the `armarx_control` package.
    ```
3. Run the example demo script below. The robot will move back and forth, then close and open both hand, and then raise
   both arms straight up 300 mm, in
   the meanwhile, the hands close gradually. This demonstrates how you can connect to the proxies on the robot and
   interact with them.
    ```shell
    cd python3-armarx
    # create the bimanual impedance controller using configuration json file
    python armarx_control/run/bimanual_imp_ctrl.py
    # see command line args
    python armarx_control/run/bimanual_imp_ctrl.py -h
   
    # Python script read the configuration json and create the bimanual impedance controller using Aron
    python armarx_control/run/bimanual_imp_ctrl.py -a
    ```
   If you see hand not closed, this might due to the issue of hand unit, which has to be activated once.
4. Run unimanual example: default using left arm, to use right arm, specify `-r`.
   ```shell
   python armarx_control/run/uni_manual_imp_ctrl.py -a -r
   ```
5. MP controller example: the robot will move the arm with a straight line trajectory encoded as Via-point Movement
   Primitive (VMP). The target is
   30 cm above the current pose. You can, e.g., move the arm to `NavigationPose` before testing the following script.
    ```shell
    cd python3-armarx
    python armarx_control/run/mp_forward_backward.py -f kinesthetic_teaching-2023-06-18-16-21-15 -t 5.0 -r -a
    # see help 
    python armarx_control/run/mp_forward_backward.py -h
    ```

More examples:

- see `armarx_control/robots/commom.py` for platform, hand and arm motion control, basic set up of robot configurations.
- see the teleoperation control in `robot-utils/robot_utils/armar/teleoperation/holistic_teleop.py`
- see `robot-policy-learning` package for movement primitive controllers.
    - task space impedance controller on both real and simulated
      robot: `robot-policy-learning/robot_policy/classical/real/taskspace_impedance_mp_controller.py`

#### Examples of Using Cameras Only

If you only need the monocular or stereo cameras, then you do not need the `ControlScenario`
which is in the `armarx_control` package.
Just do the following

```shell
from armarx_control.robots.common import Robot, cfg
robot_cfg = cfg.RobotConfig()
if cam_mode == "stereo":
    robot_cfg.stereo = cfg.StereoCameraConfig()
else:
    robot_cfg.mono = cfg.MonocularCameraConfig()
robot = Robot(robot_cfg)
if cam_mode == "stereo":
    images, info = robot.get_stereo_images()
    img = images[0]
else:
    img, depth, info = robot.get_mono_images(True)
```

You can find more examples of vision algorithms in the
[`robot-vision` package]([https://git.h2t.iar.kit.edu/sw/machine-learning-control/robot-vision])
or contact jianfeng.gao@kit.edu for more details.

### ARMAR-7 Control

To use with ARMAR-7, make sure you adapt the proxy names in the ControllerCreator component,
e.g. `RobotUnitName = Armar7Unit`, etc.

TODO:

- [ ] Since ARMAR-7 use a different hand controller, therefore, you have to comment out the code related to hand
  control in the python bindings, this must be fixed in the future. e.g. `self._load_hand_controller()` in
  `armarx_control/robots/common.py`.

Here are some instructions of how to use Python3 to control ARMAR-7

#### Running Compliance Controller

Here we provide an example of "mixed impedance velocity" controller for the right arm, since ARMAR-7 has a wrist without
torque control mode yet.
!!! IMPORTANT !!! Left arm is not calibrated yet, please do NOT use the LEFT arm in compliance mode.

By running the following scripts, you observe the right arm moving above ~10 cm from its current pose.
See terminal for output and interaction.
After that, the
arm controller will continue running at the background, therefore, you can try to perturb the arm to fine tune the
compliance parameters. Default parameters for right arm are located at
`<workspace>/armarx/skills/control/data/armarx_control/controller_config/NJointTaskspaceMixedImpedanceVelocityController/default_right_a7.json`

```shell
cd python3-armarx/armarx_control/run
python unimanual_mixed_imp_vel_ctrl_a7.py -a
```

#### Using Movement Primitives

By running the following scripts, you see the right arm following a trajectory moving its arm above a table and placing
it back to home pose. It contains a forward and backward motion. See terminal for output and interaction.

```shell
cd python3-armarx/armarx_control/run
python mixed_impedance_velocity_mp_forward_backward_a7.py -f kinesthetic_teaching-2023-06-18-16-21-15 -t 6.0 -a -r
```

#### Kinesthetic teaching with ARMAR-7

Dependencies:

- start the `controller_creator` component in the ControlScenario locally on your PC. Make sure the RobotUnitName is set
  to `Armar7Unit`.

The default arm to perform the demonstration is set to `RightArmWithoutWrist`, since at the moment
we do not have compliance mode (zero-torque control mode) in the wrist actuators, and the left arm
is not yet calibrated. Therefore, DO NOT use left arm for this script.

The recorded trajectory will be stored locally in the `armarx/skills/control` package. The python script
automatically copy the csv files to the robot RT PC for reproduction. The recordings include 4 files,

- forward in task space
- forward in joint space
- backward in task space
- backward in joint space

Follow instructions in the terminal.

1. Trigger the teaching by pressing the force torque sensor
2. move the arm as you wish
3. stop the motion and hold the arm for a second, controller and recording will stop
4. Go back to terminal to trigger reproduction:
    1. it first use the backward motion to move the arm to where it started;
    2. then a forward motion of the demonstrated motion;
    3. and finally a backward motion again to go back.

```shell
cd python3-armarx/armarx_control/run
python kinesthetic_teaching_a7.py -t 10.0 -n test_a7_kt
```

Since the `RightArmWithoutWrist` only has 6 DoF and the execution we used a mixed impedance and velocity controller
whcih
assumes the robot arm with 8 DoF. Therefore, you need to convert the joint space recordings using the following script

Hard code the trajectory you need to convert in the script.

```shell
cd python3-armarx/armarx_control/run
python adaptRecordingJSTrajectories.py 
```

#### Load Dishwasher with ARMAR-7

The aron types are defined in `armarx/skills/manipulation` package
in https://git.h2t.iar.kit.edu/sw/armarx/skills/manipulation/-/tree/main/source/armarx/manipulation/skills/aron/mp_skill?ref_type=heads.

Including the action names in `ActionNames.xml`, which serves as enum to choose how target pose, via pose are derived
in [ReachTcpPoseRelativeToObjectWithMP skill](https://git.h2t.iar.kit.edu/sw/armarx/skills/manipulation/-/blob/main/source/armarx/manipulation/skills/ReachTcpPoseRelativeToObjectWithMP.h?ref_type=heads)
and which motion trajecotries are used to train mp
in [ExecuteTraj skill](https://git.h2t.iar.kit.edu/sw/armarx-integration/robots/armar7/skills/-/tree/main/source/armar7/skills/mp_skill?ref_type=heads).

After kinesthetic teaching, and converting the joint space trajectories, you can define a list a trajectories in the
ExecuteTraj skill in the map, with a key value corresponding to a action name defined in the `ActionNames.xml`.
Similarly, define a function in `ReachTcpPoseRelativeToObjectWithMP` skill (e.g. like `parseParamPushDownDishWasher`) to
derive the target pose, via pose, and hand shape either based on the articulated object or the robot root depending on
the task.

To enable safe interaction (i.e. let the NJoint controller stop when large force is detected), you can
set `param.forceTorqueGuardCanVal = 0.1;` in `parseParamPushDownDishWasher` function, which is a parameter for the
ExecuteTraj skill. The value corresponds to the canonical value of VMP, where the force torque sensor is calibrated and
zeroed, after that point, the safe guard based on FT sensor is enabled.

Note that, all target pose, via pose need to be defined in the robot root frame. while the visualization need to be in
global frame.

##### Potential improvement

- [ ] move to dishwasher handle, the motion is fast before contacting the door, this can be modified by adjusting the
  via pose canonical value, e.g. change `viaFromHandle.canVal = 0.3;` to some larger value 0.5 (to let this viapoint
  happen ealier, therefore the arm has more time to make contact thus
  slower) https://git.h2t.iar.kit.edu/sw/armarx/skills/manipulation/-/blob/main/source/armarx/manipulation/skills/ReachTcpPoseRelativeToObjectWithMP.cpp?ref_type=heads#L297

- [ ] the put object in dishwasher: goal pose is too low, therefore, the hand push the object to the basket to much,
  adjust the z dir offset `targetPosePlaceObject.block<3, 1>(0, 3) += Eigen::Vector3f{0.0, 200., 30.};`
  in https://git.h2t.iar.kit.edu/sw/armarx/skills/manipulation/-/blob/main/source/armarx/manipulation/skills/ReachTcpPoseRelativeToObjectWithMP.cpp?ref_type=heads#L831

- [ ] As requested, to split
  the [large integration skill](https://git.h2t.iar.kit.edu/sw/armarx-integration/robots/armar7/skills/-/blob/main/source/armar7/skills/manipulation/skills/LoadDishwasher.cpp?ref_type=heads):
  open dishwasher (to the point of finishing the pull out basket); put object in dishwasher (may have multiple modes:
  navigate to contertop and grasp, human handover, or object already in hand, we have to discuss); and finally close
  door.